/*
 * serialmon-ioctl.c
 *
 * Copyright (C) 2001 Peter Fox
 *
 * Module to Intercept ioctls for a serial port to add the TIOCSERGSTATS ioctl for serialmon
 *
 * Use ioctl TIOCSERRELMOD to remove the interception and allow the module to be removed.
 *
 * Specify the device to intercept using the dev module parameter
 *
 * This is somewhat more complicated than we'd like because the driver
 * structure pointed to by the file structure is just a copy, and we
 * need to get at the real one.
 *
 * We also have to check to see if the driver we are intercepting is
 * in a loadable module, and increase it's usage count to prevent it
 * getting removed before us.
 *
 *
 */

#define _LOOSE_KERNEL_NAMES /* For glibc (see features.h) */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kdev_t.h>
#include <linux/tty.h>
#include <asm/uaccess.h>
#include "serialmon-ioctl.h"

#define PROTECT_REMOVAL

MODULE_AUTHOR("Peter Fox");
MODULE_DESCRIPTION("Add TIOCSERGSTATS ioctl to a serial driver");
EXPORT_NO_SYMBOLS;
MODULE_PARM(dev, "s");
MODULE_PARM(verbose, "i");

extern struct file *filp_open(const char *, int, int);
extern int filp_close(struct file *, fl_owner_t id);

static char *dev = "/dev/ttyS0"; /* default */
static int verbose = 0; /* You really don't want to use this.. */

static struct module *driver_module;

struct tty_driver *driver; /* Have to remember this to un-chain my ioctl handler */

static int (*orig_ioctl)(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg);

/* Find the module containing the driver we have hooked */

static struct module *
addr2mod(void *ptr)
{
	struct module *mptr = &__this_module; /* Start from where we are */
	while(mptr)
	{
		if((ptr > (void *)mptr)&&(ptr < (void *)mptr + mptr->size))
			return mptr;
		mptr = mptr->next;
	}
	return 0;
}

/* From serial.c */

#define SERIAL_PARANOIA_CHECK

static inline int
serial_paranoia_check(struct async_struct *info, kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
    static const char *badmagic =
        "Warning: bad magic number for serial struct (%s) in %s\n";
    static const char *badinfo =
        "Warning: null async_struct for (%s) in %s\n";

    if (!info) {
        printk(badinfo, kdevname(device), routine);
        return 1;
    }
    if (info->magic != SERIAL_MAGIC) {
        printk(badmagic, kdevname(device), routine);
        return 1;
    }
#endif
    return 0;
}

/* My extra ioctls */

static int
serialmon_ioctl(struct tty_struct *tty, struct file * file,	unsigned int cmd, unsigned long arg)
{
    struct async_struct * info = (struct async_struct *)tty->driver_data;
    struct serial_stats_struct k_stats;
    struct serial_stats_struct *retinfo = (struct serial_stats_struct *)arg;
    struct async_icount *icount;
	unsigned long flags;

    if(serial_paranoia_check(info, tty->device, "serialmon_ioctl"))
        return -ENODEV;

/*	printk("serialmon-ioctl: cmd = %08x\n", cmd); */

	if(cmd == TIOCSERRELMOD)
	{
		int retval = 0;
		printk("original ptr = %p, new ptr = %p, my ioctl at %p\n", orig_ioctl, driver->ioctl, &serialmon_ioctl);
		save_flags(flags);
		cli();
		if((driver->ioctl == &serialmon_ioctl)&&(tty->driver.ioctl == &serialmon_ioctl)&&(*(driver->refcount) == 1))
		{
			if(verbose) printk("serialmon-ioctl: Removing TIOCSERGSTATS ioctl for %s\n", dev);
			driver->ioctl = orig_ioctl;
			tty->driver.ioctl = orig_ioctl; /* Just in case someone else tries to use the device */
			printk("serialmon-ioctl: Removed TIOCSERGSTATS ioctl for %s\n", dev);
			if(driver_module)
			{
				printk("serialmon-ioctl: Decrementing usage of driver module '%s'\n", driver_module->name);
				__MOD_DEC_USE_COUNT(driver_module);
			}
#ifdef PROTECT_REMOVAL
			MOD_DEC_USE_COUNT;
#endif
		}
		else
		{
			printk("serialmon-ioctl: Cannot remove TIOCSERGSTATS ioctl for %s\n", dev);
			retval = -EBUSY;
		}
		restore_flags(flags);
		return retval;
	}

    if(cmd != TIOCSERGSTATS)
		return (*orig_ioctl)(tty, file, cmd, arg);

    icount = &info->state->icount;
	if(verbose) printk("Got TIOCSERGSTATS ioctl call\n");
	save_flags(flags);
    cli();
    do_gettimeofday(&k_stats.ts);
    k_stats.count = info->tty->count;
    k_stats.rx = icount->rx;
    k_stats.tx = icount->tx;
	restore_flags(flags);
    if (copy_to_user(retinfo,&k_stats,sizeof(*retinfo)))
        return -EFAULT;
    return 0;

}

int
init_module(void)
{
    struct file *filp = filp_open(dev, O_EXCL, 0);
	struct tty_struct *ttyp; 
    struct tty_driver *driver_copy;
    struct tty_driver *p;
	int retval;

	if(IS_ERR(filp))
	{
		printk("Cannot open %s - error %ld\n", dev, PTR_ERR(filp));
		return PTR_ERR(filp);
	}
	printk("serialmon-ioctl: Installing TIOCSERGSTATS ioctl for %s..\n", dev);
	ttyp = (struct tty_struct *)filp->private_data;
	if((!ttyp)||(ttyp->magic != TTY_MAGIC))
	{
		printk("serialmon-ioctl: %s is not a TTY\n", dev);
		retval = -ENOTTY;
		goto bailout;
	}
	if((ttyp->driver.magic != TTY_DRIVER_MAGIC))
	{
		printk("serialmon-ioctl: %s does not have a TTY_DRIVER\n", dev);
		retval = -ENODEV;
		goto bailout;
	}
	driver_copy = &ttyp->driver;
	if(((p = ttyp->driver.prev) && (p = p->next)))
	{
		printk("serialmon-ioctl: found real driver structure at (prev next) %p\n", p);
	}
	else if(((p = ttyp->driver.next) && (p = p->prev)))
	{
		printk("serialmon-ioctl: found real driver structure at (next prev) %p\n", p);
	}
	else
	{
		printk("serialmon-ioctl: Cannot locate driver, not enough tty drivers loaded\n");
		retval = -ENODEV;
		goto bailout;
	}
	if((p->magic != TTY_DRIVER_MAGIC))
	{
		printk("serialmon-ioctl: %s does not have a TTY_DRIVER\n", dev);
		retval = -ENODEV;
		goto bailout;
	}
	driver = p;
#if 1
	/*
	 * Install our ioctl handler to the real driver
	 */
	if(driver->ioctl == driver_copy->ioctl)
	{
		orig_ioctl = driver->ioctl;
		printk("original ptr = %p, new ptr = %p, copy_ptr = %p, my ioctl at %p\n", orig_ioctl, driver->ioctl, driver_copy->ioctl, &serialmon_ioctl);
		driver->ioctl = &serialmon_ioctl;
		driver_copy->ioctl = &serialmon_ioctl;
		driver_module = addr2mod(orig_ioctl);
		if(driver_module)
		{
			printk("serialmon-ioctl: Incrementing usage of driver module '%s'\n", driver_module->name);
			__MOD_INC_USE_COUNT(driver_module);
		}
#ifdef PROTECT_REMOVAL
		MOD_INC_USE_COUNT;
#endif
		printk("serialmon-ioctl: Installed TIOCSERGSTATS ioctl for %s\n", dev);
#endif
	}
	else
	{
		printk("serialmon-ioctl: ioctl addresses don't match, can't install ioctl intercept.\n");
		retval = -EBUSY;
		goto bailout;
	}
	filp_close(filp, 0);
	return 0;

 bailout:
	filp_close(filp, 0);
	return retval;
}

void
cleanup_module(void)
{
	/* Nothing to do, as the TIOCSERRELMOD ioctl must have already done the work */
#if 0 /* Should be 0 */
	/* Have to open the file to ensure the pointers are good */
    struct file *filp = filp_open(dev, O_EXCL, 0);
	struct tty_struct *ttyp; 
    struct tty_driver *driver_copy;

	printk("serialmon-ioctl: Removing TIOCSERGSTATS ioctl for %s..\n", dev);
	if(IS_ERR(filp))
	{
		printk("Cannot open %s - error %ld\n", dev, PTR_ERR(filp));
		driver->ioctl = orig_ioctl;
		return;
	}
	printk("serialmon-ioctl: Installing TIOCSERGSTATS ioctl for %s..\n", dev);
	ttyp = (struct tty_struct *)filp->private_data;
	if((!ttyp)||(ttyp->magic != TTY_MAGIC))
	{
		printk("serialmon-ioctl: %s is not a TTY\n", dev);
		goto dominimum;
	}
	if((ttyp->driver.magic != TTY_DRIVER_MAGIC))
	{
		printk("serialmon-ioctl: %s does not have a TTY_DRIVER\n", dev);
		goto dominimum;
	}
	driver_copy = &ttyp->driver;
	printk("original ptr = %p, new ptr = %p, copy_ptr = %p, my ioctl at %p\n", orig_ioctl, driver->ioctl, driver_copy->ioctl, &serialmon_ioctl);
	if(verbose) printk("serialmon-ioctl: Removing TIOCSERGSTATS ioctl for %s\n", dev);
#if 0
	if(driver->ioctl != &serialmon_ioctl)
	{
		printk("original ptr = %p, new ptr = %p, my ioctl at %p\n", orig_ioctl, driver->ioctl, &serialmon_ioctl);
		panic("serialmon-ioctl pointers screwed up");
	}
#endif
	driver->ioctl = orig_ioctl;
	driver_copy->ioctl = orig_ioctl;
	printk("serialmon-ioctl: Removed TIOCSERGSTATS ioctl for %s\n", dev);
	filp_close(filp, 0);
	return;

 dominimum:
	driver->ioctl = orig_ioctl;
	filp_close(filp, 0);
	return;
#endif
}

