#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <sys/time.h>	/* struct timeval */
#include <signal.h>

#include "serialbw.h"
#include "shmemserv.h"

#define DEBUG
#undef DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

#define MAX_Q_ENTRIES 50

typedef struct queue
{
	pid_t data[MAX_Q_ENTRIES];
	int len;
	int rptr;
	int wptr;
} QUEUE, *QUEUE_PTR;

static QUEUE rxq, txq;

#define q_init(q)	(void)memset((void *)(q), 0, sizeof(QUEUE))
#define q_room(q)	((q)->len < MAX_Q_ENTRIES)
#define q_avail(q)	((q)->len)
#define q_write(q, d) ((q)->data[(q)->wptr++] = (d), (q)->wptr = (q)->wptr < MAX_Q_ENTRIES ? (q)->wptr : 0, (q)->len++)
static inline int q_read(QUEUE_PTR q)
{
	int t = (q)->data[(q)->rptr++];
	(q)->rptr = (q)->rptr < MAX_Q_ENTRIES ? (q)->rptr : 0;
	(q)->len--;
	return t;
}

static struct timeval last_sig;	/* Sent last signal */
static struct timeval prevstat;	/* Previous statistics time; */
static int prev_rx, prev_tx;	/* Char counts */
static int can_rx;
static int can_tx;
static int reg_rx, reg_tx;	/* Processes being registered */

double rate_threshold = 1000.0;	/* Chars per sec below which can start another process */
double min_interval = 0.5;		/* Min interval between starting processes */

extern double usecdiff(struct timeval *old, struct timeval *new);

int
shmemserv_init(char *line, SERIALBW_PTR *sbwptr)
{
	int shmid;
	SERIALBW_PTR ptr;
	key_t devkey = ftok(line, 'S');
	if(devkey == (key_t)(-1))
	{
		return SHMEM_FAILED_KEY;
	}
#ifdef DEBUG
	printf("key is %d\n", devkey);
#endif
	/* Make shared segment */
	shmid = shmget(devkey, SEGSIZE, IPC_CREAT | 0666);
	if(shmid == -1)
	{
		return SHMEM_FAILED_GET;
	}
#ifdef DEBUG
	printf("shmid is %d\n", shmid);
#endif
	ptr = (SERIALBW_PTR)shmat(shmid, 0, 0);
	if(ptr == (SERIALBW_PTR)(-1))
	{
		return SHMEM_FAILED_MAP;
	}
	(void)memset((void *)ptr, 0, sizeof(SERIALBW));
	*sbwptr = ptr;
	/* Make sure ok when shmemserv_linedown called before shmemserv_lineup */
	q_init(&rxq);
	q_init(&txq);
	return SHMEM_INIT_OK;
}

/*
 * Get called when line goes up
 */
void
shmemserv_lineup(SERIALBW_PTR ptr, struct timeval *now)
{
#ifdef DEBUG
	printf("shmemserv_lineup\n");
#endif
	if(ptr)
	{
		(void)memset((void *)ptr, 0, sizeof(SERIALBW));
		ptr->sbw_flags = SBW_SERVEROK | SBW_LINEUP;
	}
	q_init(&rxq);
	q_init(&txq);
	prevstat = last_sig = *now;
	reg_rx = 0;
	reg_tx = 0;
}

/*
 * Get called at startup, and when line goes down
 */
void
shmemserv_linedown(SERIALBW_PTR ptr, struct timeval *now)
{
#ifdef DEBUG
	printf("shmemserv_linedown\n");
#endif
	if(ptr)
	{
		(void)memset((void *)ptr, 0, sizeof(SERIALBW));
		ptr->sbw_flags = SBW_SERVEROK;
	}
	/* Kick off all the waiting processes if available */
	while(q_avail(&rxq))
	{
		int pid = q_read(&rxq);
#ifdef DEBUG
		printf("Kicking rx process %d\n", pid);
#endif
		kill(pid, SIGCONT);
	}
	while(q_avail(&txq))
	{
		int pid = q_read(&txq);
#ifdef DEBUG
		printf("Kicking tx process %d\n", pid);
#endif
		kill(pid, SIGCONT);
	}
}

/*
 * Get called when exiting
 */
void
shmemserv_shutdown(SERIALBW_PTR ptr, struct timeval *now)
{
#ifdef DEBUG
	printf("shmemserv_shutdown\n");
#endif
	shmemserv_linedown(ptr, now);
	if(ptr)
	{
		ptr->sbw_flags = 0;
	}
}

/*
 * Get called at regular intervals, do protocol
 */
void
shmemserv_scan(SERIALBW_PTR ptr, struct timeval *now)
{
	double tdiff;
	/* If shmem access failed, forget it */
	if(!ptr)
		return;
	/* Check for new settings */
	if(ptr->new_rate)
	{
		rate_threshold = ptr->new_rate;
		ptr->new_rate = 0.0;
	}
	if(ptr->new_tgap)
	{
		min_interval = ptr->new_tgap;
		ptr->new_tgap = 0.0;
	}
	/* If line not up can stop now. */
	if((ptr->sbw_flags & SBW_LINEUP) == 0)
		return;
	/* Check for requestors */
	if((!reg_rx)&&(ptr->rx_req)&&(q_room(&rxq)))
	{
		reg_rx = ptr->rx_req;
#ifdef DEBUG
		printf("Registering rx process %d\n", reg_rx);
#endif
		q_write(&rxq, reg_rx);
		ptr->rx_ack = reg_rx;
	}
	else if(ptr->rx_ack == 0)
	{
		reg_rx = 0;
	}
	if((!reg_tx)&&(ptr->tx_req)&&(q_room(&txq)))
	{
		reg_tx = ptr->tx_req;
#ifdef DEBUG
		printf("Registering tx process %d\n", reg_tx);
#endif
		q_write(&txq, reg_tx);
		ptr->tx_ack = reg_tx;
	}
	else if(ptr->tx_ack == 0)
	{
		reg_tx = 0;
	}
	/* Keep track of saturation, but wait min interval before looking */
	if((tdiff = usecdiff(&prevstat, now)) > 0.0)
	{
		/* Don't calculate if too soon, or already set both */
		if((usecdiff(&last_sig, now) > min_interval)&&((can_rx & can_tx) == 0))
		{
			double rx_rate, tx_rate;
			rx_rate = (ptr->rx_count - prev_rx)/tdiff;
			tx_rate = (ptr->tx_count - prev_tx)/tdiff;
			if(rx_rate < rate_threshold)
				can_rx |= 1;
			if(tx_rate < rate_threshold)
				can_tx |= 1;
		}
		prev_rx = ptr->rx_count;
		prev_tx = ptr->tx_count;
		prevstat = *now;
	}
	/* Kick off process if one available */
	if(can_rx && q_avail(&rxq))
	{
		int pid = q_read(&rxq);
#ifdef DEBUG
		printf("Kicking rx process %d\n", pid);
#endif
		ptr->rx_go = pid;
		kill(pid, SIGCONT);
		last_sig = *now;
		can_rx = 0;
	}
	if(can_tx && q_avail(&txq))
	{
		int pid = q_read(&txq);
#ifdef DEBUG
		printf("Kicking tx process %d\n", pid);
#endif
		ptr->tx_go = pid;
		kill(pid, SIGCONT);
		last_sig = *now;
		can_tx = 0;
	}
}
